! WHIZARD 2.2.0 alpha Sep 12 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module beam_structures
  
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use unit_tests
  use variables

  implicit none
  private

  public :: beam_structure_t
  public :: beam_structures_test

  type :: beam_structure_entry_t
     logical :: is_valid = .false.
     type(string_t) :: name
   contains
     procedure :: to_string => beam_structure_entry_to_string
  end type beam_structure_entry_t
  
  type :: beam_structure_record_t
     type(beam_structure_entry_t), dimension(:), allocatable :: entry
  end type beam_structure_record_t

  type :: beam_structure_t
     type(beam_structure_record_t), dimension(:), allocatable :: record
     type(var_list_t) :: var_list
   contains
     procedure :: write => beam_structure_write
     procedure :: to_string => beam_structure_to_string
     procedure :: init => beam_structure_init
     procedure :: set_entry => beam_structure_set_entry
     procedure :: expand => beam_structure_expand
     procedure :: get_n_record => beam_structure_get_n_record
     procedure :: get_i_entry => beam_structure_get_i_entry
     procedure :: get_name => beam_structure_get_name
  end type beam_structure_t
  

  abstract interface
     function strfun_mode_fun (name) result (n)
       import
       type(string_t), intent(in) :: name
       integer :: n
     end function strfun_mode_fun
  end interface
  

contains
  
  function beam_structure_entry_to_string (object) result (string)
    class(beam_structure_entry_t), intent(in) :: object
    type(string_t) :: string
    if (object%is_valid) then
       string = object%name
    else
       string = "none"
    end if
  end function beam_structure_entry_to_string

  subroutine beam_structure_write (object, unit)
    class(beam_structure_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    if (allocated (object%record)) then
       if (size (object%record) > 0) then
          write (u, "(1x,A)")  "Beam structure:"
          write (u, "(3x,A)")  char (object%to_string ())
       else
          write (u, "(1x,A)")  "Beam structure: [trivial]"
       end if
    else
       write (u, "(1x,A)")  "Beam structure: [undefined]"
    end if
  end subroutine beam_structure_write
  
  function beam_structure_to_string (object) result (string)
    class(beam_structure_t), intent(in) :: object
    type(string_t) :: string
    integer :: i, j
    if (allocated (object%record)) then
       string = ""
       if (size (object%record) > 0) then
          do i = 1, size (object%record)
             if (i > 1)  string = string // " => "
             do j = 1, size (object%record(i)%entry)
                if (j > 1)  string = string // ", "
                string = string // object%record(i)%entry(j)%to_string ()
             end do
          end do
       else
          string = "[trivial]"
       end if
    else
       string = "[undefined]"
    end if
  end function beam_structure_to_string
    
  subroutine beam_structure_init (beam_structure, dim_array, var_list)
    class(beam_structure_t), intent(out) :: beam_structure
    integer, dimension(:), intent(in) :: dim_array
    type(var_list_t), intent(in) :: var_list
    integer :: i
    allocate (beam_structure%record (size (dim_array)))
    do i = 1, size (dim_array)
       allocate (beam_structure%record(i)%entry (dim_array(i)))
    end do
    call var_list_init_snapshot (beam_structure%var_list, var_list)
  end subroutine beam_structure_init
    
  subroutine beam_structure_set_entry (beam_structure, i, j, name)
    class(beam_structure_t), intent(inout) :: beam_structure
    integer, intent(in) :: i, j
    type(string_t), intent(in) :: name
    associate (entry => beam_structure%record(i)%entry(j))
      entry%name = name
      entry%is_valid = .true.
    end associate
  end subroutine beam_structure_set_entry

  subroutine beam_structure_expand (beam_structure, strfun_mode)
    class(beam_structure_t), intent(inout) :: beam_structure
    procedure(strfun_mode_fun) :: strfun_mode
    type(beam_structure_record_t), dimension(:), allocatable :: new
    integer :: n_record, i, j
    if (.not. allocated (beam_structure%record))  return
    do i = 1, size (beam_structure%record)
       associate (entry => beam_structure%record(i)%entry)
         do j = 1, size (entry)
            select case (strfun_mode (entry(j)%name))
            case (0);  entry(j)%is_valid = .false.
            end select
         end do
       end associate
    end do
    n_record = 0
    do i = 1, size (beam_structure%record)
       associate (entry => beam_structure%record(i)%entry)
         select case (size (entry))
         case (1)
            if (entry(1)%is_valid) then
               select case (strfun_mode (entry(1)%name))
               case (1);  n_record = n_record + 2
               case (2);  n_record = n_record + 1
               end select
            end if
         case (2)
            do j = 1, 2
               if (entry(j)%is_valid) then
                  select case (strfun_mode (entry(j)%name))
                  case (1);  n_record = n_record + 1
                  case (2)
                     call beam_structure%write ()
                     call msg_fatal ("Pair spectrum used as &
                          &single-particle structure function")
                  end select
               end if
            end do
         end select
       end associate
    end do
    allocate (new (n_record))
    n_record = 0
    do i = 1, size (beam_structure%record)
       associate (entry => beam_structure%record(i)%entry)
         select case (size (entry))
         case (1)
            if (entry(1)%is_valid) then
               select case (strfun_mode (entry(1)%name))
               case (1)
                  n_record = n_record + 1
                  allocate (new(n_record)%entry (2))
                  new(n_record)%entry(1) = entry(1)
                  n_record = n_record + 1
                  allocate (new(n_record)%entry (2))
                  new(n_record)%entry(2) = entry(1)
               case (2)
                  n_record = n_record + 1
                  allocate (new(n_record)%entry (1))
                  new(n_record)%entry(1) = entry(1)
               end select
            end if
         case (2)
            do j = 1, 2
               if (entry(j)%is_valid) then
                  n_record = n_record + 1
                  allocate (new(n_record)%entry (2))
                  new(n_record)%entry(j) = entry(j)
               end if
            end do
         end select
       end associate
    end do
    call move_alloc (from = new, to = beam_structure%record)
  end subroutine beam_structure_expand
    
  function beam_structure_get_n_record (beam_structure) result (n)
    class(beam_structure_t), intent(in) :: beam_structure
    integer :: n
    if (allocated (beam_structure%record)) then
       n = size (beam_structure%record)
    else
       n = 0
    end if
  end function beam_structure_get_n_record
  
  function beam_structure_get_i_entry (beam_structure, i) result (i_entry)
    class(beam_structure_t), intent(in) :: beam_structure
    integer, intent(in) :: i
    integer, dimension(:), allocatable :: i_entry
    integer :: j
    associate (record => beam_structure%record(i))
      select case (size (record%entry))
      case (1)
         if (record%entry(1)%is_valid) then
            allocate (i_entry (2), source = [1, 2])
         else
            allocate (i_entry (0))
         end if
      case (2)
         if (all (record%entry%is_valid)) then
            allocate (i_entry (2), source = [1, 2])
         else if (record%entry(1)%is_valid) then
            allocate (i_entry (1), source = [1])
         else if (record%entry(2)%is_valid) then
            allocate (i_entry (1), source = [2])
         else
            allocate (i_entry (0))
         end if
      end select
    end associate
  end function beam_structure_get_i_entry
  
  function beam_structure_get_name (beam_structure, i) result (name)
    class(beam_structure_t), intent(in) :: beam_structure
    integer, intent(in) :: i
    type(string_t) :: name
    associate (record => beam_structure%record(i))
      if (record%entry(1)%is_valid) then
         name = record%entry(1)%name
      else if (size (record%entry) == 2) then
         name = record%entry(2)%name
      end if
    end associate
  end function beam_structure_get_name
  

  subroutine beam_structures_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (beam_structures_1, "beam_structures_1", &
         "empty beam structure record", &
         u, results)
    call test (beam_structures_2, "beam_structures_2", &
         "beam structure records", &
         u, results)
    call test (beam_structures_3, "beam_structures_3", &
         "beam structure expansion", &
         u, results)
    call test (beam_structures_4, "beam_structures_4", &
         "beam structure contents", &
         u, results)
end subroutine beam_structures_test
  
  subroutine beam_structures_1 (u)
    integer, intent(in) :: u
    type(beam_structure_t) :: beam_structure
    
    write (u, "(A)")  "* Test output: beam_structures_1"
    write (u, "(A)")  "*   Purpose: display empty beam structure record"
    write (u, "(A)")

    call beam_structure%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: beam_structures_1"
    
  end subroutine beam_structures_1
  
  subroutine beam_structures_2 (u)
    integer, intent(in) :: u
    type(beam_structure_t) :: beam_structure
    integer, dimension(0) :: empty_array
    type(var_list_t) :: var_list
    
    write (u, "(A)")  "* Test output: beam_structures_2"
    write (u, "(A)")  "*   Purpose: setup beam structure records"
    write (u, "(A)")

    call var_list_append_int (var_list, var_str ("foo"), 42)

    call beam_structure%init (empty_array, var_list)
    call beam_structure%write (u)
    
    write (u, "(A)")
    write (u, "(1x,A)")  "Variables:"
    call var_list_write (beam_structure%var_list, u)
    
    write (u, "(A)")

    call beam_structure%init ([1], var_list)
    call beam_structure%set_entry (1, 1, var_str ("a"))
    call beam_structure%write (u)
    
    write (u, "(A)")

    call beam_structure%init ([2], var_list)
    call beam_structure%set_entry (1, 1, var_str ("a"))
    call beam_structure%set_entry (1, 2, var_str ("b"))
    call beam_structure%write (u)
    
    write (u, "(A)")

    call beam_structure%init ([2, 1], var_list)
    call beam_structure%set_entry (1, 1, var_str ("a"))
    call beam_structure%set_entry (1, 2, var_str ("b"))
    call beam_structure%set_entry (2, 1, var_str ("c"))
    call beam_structure%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Test output end: beam_structures_2"
    
  end subroutine beam_structures_2
  
  function test_strfun_mode (name) result (n)
    type(string_t), intent(in) :: name
    integer :: n
    select case (char (name))
    case ("a");  n = 2
    case ("b");  n = 1
    case default;  n = 0
    end select
  end function test_strfun_mode

  subroutine beam_structures_3 (u)
    integer, intent(in) :: u
    type(beam_structure_t) :: beam_structure
    type(var_list_t) :: var_list
    
    write (u, "(A)")  "* Test output: beam_structures_3"
    write (u, "(A)")  "*   Purpose: expand beam structure records"
    write (u, "(A)")

    write (u, "(A)")  "* Pair spectrum (keep as-is)"
    write (u, "(A)")

    call beam_structure%init ([1], var_list)
    call beam_structure%set_entry (1, 1, var_str ("a"))
    call beam_structure%write (u)

    write (u, "(A)")

    call beam_structure%expand (test_strfun_mode)
    call beam_structure%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Structure function pair (expand)"
    write (u, "(A)")

    call beam_structure%init ([2], var_list)
    call beam_structure%set_entry (1, 1, var_str ("b"))
    call beam_structure%set_entry (1, 2, var_str ("b"))
    call beam_structure%write (u)

    write (u, "(A)")

    call beam_structure%expand (test_strfun_mode)
    call beam_structure%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Structure function (separate and expand)"
    write (u, "(A)")

    call beam_structure%init ([1], var_list)
    call beam_structure%set_entry (1, 1, var_str ("b"))
    call beam_structure%write (u)

    write (u, "(A)")

    call beam_structure%expand (test_strfun_mode)
    call beam_structure%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Combination"
    write (u, "(A)")

    call beam_structure%init ([1, 1], var_list)
    call beam_structure%set_entry (1, 1, var_str ("a"))
    call beam_structure%set_entry (2, 1, var_str ("b"))
    call beam_structure%write (u)

    write (u, "(A)")

    call beam_structure%expand (test_strfun_mode)
    call beam_structure%write (u)
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: beam_structures_3"
    
  end subroutine beam_structures_3
  
  subroutine beam_structures_4 (u)
    integer, intent(in) :: u
    type(beam_structure_t) :: beam_structure
    type(var_list_t) :: var_list
    integer :: i
    
    write (u, "(A)")  "* Test output: beam_structures_4"
    write (u, "(A)")  "*   Purpose: check the API"
    write (u, "(A)")

    write (u, "(A)")  "* Structure-function combination"
    write (u, "(A)")

    call beam_structure%init ([1, 2, 2], var_list)
    call beam_structure%set_entry (1, 1, var_str ("a"))
    call beam_structure%set_entry (2, 1, var_str ("b"))
    call beam_structure%set_entry (3, 2, var_str ("c"))
    call beam_structure%write (u)

    write (u, "(A)")
    write (u, "(1x,A,I0)")  "n_record = ", beam_structure%get_n_record ()
    
    do i = 1, 3
       write (u, "(A)")
       write (u, "(1x,A,I0,A,A)")  "name(", i, ") = ", &
            char (beam_structure%get_name (i))
       write (u, "(1x,A,I0,A,2(1x,I0))")  "i_entry(", i, ") =", &
            beam_structure%get_i_entry (i)
    end do
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: beam_structures_4"
    
  end subroutine beam_structures_4
  

end module beam_structures
