! WHIZARD 2.2.0 alpha Sep 12 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module dispatch
  
  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use constants, only: PI !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use os_interface
  use unit_tests
  use variables
  use flavors
  use interactions
  use models
  use process_constants
  use sf_mappings
  use sf_base
  use sf_pdf_builtin
  use sf_lhapdf
  use sf_circe1
  use sf_circe2
  use sf_lumi_linker
  use sf_isr
  use sf_epa
  use sf_ewa
  use sf_escan
  use sf_beam_events
  use sf_user
  use rng_base
  use rng_tao
  use mci_base
  use mci_midpoint
  use mci_vamp
  use phs_base
  use mappings
  use phs_forests
  use phs_single
  use phs_wood
  use prc_core_def
  use prc_test
  use beams
  use prc_omega
  use prc_core
  use processes
  use beam_structures
  use eio_base
  use eio_raw
  use eio_lhef
  use eio_weights
  use rt_data
  
  implicit none
  private

  public :: dispatch_core_def
  public :: dispatch_core
  public :: dispatch_core_update
  public :: dispatch_core_restore
  public :: dispatch_mci
  public :: dispatch_phs
  public :: dispatch_rng
  public :: dispatch_sf_data
  public :: dispatch_sf_config
  public :: dispatch_eio
  public :: dispatch_test

contains
  
  subroutine dispatch_core_def (core_def, prt_in, prt_out, global)
    class(prc_core_def_t), allocatable, intent(inout) :: core_def
    type(string_t), dimension(:), intent(in) :: prt_in
    type(string_t), dimension(:), intent(in) :: prt_out
    type(rt_data_t), intent(in) :: global
    type(string_t) :: method
    type(string_t) :: model_name
    type(string_t) :: restrictions
    logical :: openmp_support
    logical :: report_progress
    type(string_t) :: extra_options
    associate (var_list => global%var_list)
      method = var_list_get_sval (var_list, var_str ("$method"))
      model_name = var_list_get_sval (var_list, var_str ("$model_name"))
      select case (char (method))
      case ("unit_test")
         allocate (prc_test_def_t :: core_def)
         select type (core_def)
         type is (prc_test_def_t)
            call core_def%init (model_name, prt_in, prt_out)
         end select
      case ("omega")
         restrictions = var_list_get_sval (var_list, &
              var_str ("$restrictions"))
         openmp_support = var_list_get_lval (var_list, &
              var_str ("?omega_openmp"))
         report_progress = var_list_get_lval (var_list, &
              var_str ("?report_progress"))
         extra_options = var_list_get_sval (var_list, &
              var_str ("$omega_flags"))
         allocate (omega_def_t :: core_def)
         select type (core_def)
         type is (omega_def_t)
            call core_def%init (model_name, prt_in, prt_out, &
                 restrictions, openmp_support, report_progress, extra_options)
         end select
      case default
         call msg_fatal ("Process configuration: method '" &
              // char (method) // "' not implemented")
      end select
    end associate
  end subroutine dispatch_core_def
    
  subroutine dispatch_core (core, core_def, model, helicity_selection)
    class(prc_core_t), allocatable, intent(inout) :: core
    class(prc_core_def_t), intent(in) :: core_def
    type(model_t), intent(in), target, optional :: model
    type(helicity_selection_t), intent(in), optional :: helicity_selection
    select type (core_def)
    type is (prc_test_def_t)
       allocate (test_t :: core)
    type is (omega_def_t)
       allocate (prc_omega_t :: core)
       select type (core)
       type is (prc_omega_t)
          call core%set_parameters (model, helicity_selection)
       end select
    class default
       call msg_bug ("Process core: unexpected process definition type")
    end select
  end subroutine dispatch_core

  subroutine dispatch_core_update (core, model, helicity_selection, saved_core)
    class(prc_core_t), allocatable, intent(inout) :: core
    type(model_t), intent(in), optional, target :: model
    type(helicity_selection_t), intent(in), optional :: helicity_selection
    class(prc_core_t), allocatable, intent(inout), optional :: saved_core
    if (present (saved_core)) then
       allocate (saved_core, source = core)
    end if
    select type (core)
    type is (test_t)
    type is (prc_omega_t)
       call core%set_parameters (model, helicity_selection)
       call core%activate_parameters ()
    class default
       call msg_bug ("Process core update: unexpected process definition type")
    end select
  end subroutine dispatch_core_update

  subroutine dispatch_core_restore (core, saved_core)
    class(prc_core_t), allocatable, intent(inout) :: core
    class(prc_core_t), allocatable, intent(inout) :: saved_core
    call move_alloc (from = saved_core, to = core)
    select type (core)
    type is (test_t)
    type is (prc_omega_t)
       call core%activate_parameters ()
    class default
       call msg_bug ("Process core restore: unexpected process definition type")
    end select
  end subroutine dispatch_core_restore

  subroutine dispatch_mci (mci, global, process_id)
    class(mci_t), allocatable, intent(inout) :: mci
    type(rt_data_t), intent(in) :: global
    type(string_t), intent(in) :: process_id
    type(string_t) :: run_id
    type(string_t) :: integration_method
    type(grid_parameters_t) :: grid_par
    logical :: rebuild_grids
    integration_method = var_list_get_sval (global%var_list, &
         var_str ("$integration_method"))
    select case (char (integration_method))
    case ("midpoint")
       allocate (mci_midpoint_t :: mci)
    case ("vamp", "default")
       associate (var_list => global%var_list)
         grid_par%threshold_calls = &
              var_list_get_ival (var_list, var_str ("threshold_calls"))
         grid_par%min_calls_per_channel = &
              var_list_get_ival (var_list, var_str ("min_calls_per_channel"))
         grid_par%min_calls_per_bin = &
              var_list_get_ival (var_list, var_str ("min_calls_per_bin"))
         grid_par%min_bins = &
              var_list_get_ival (var_list, var_str ("min_bins"))
         grid_par%max_bins = &
              var_list_get_ival (var_list, var_str ("max_bins"))
         grid_par%stratified = &
              var_list_get_lval (var_list, var_str ("?stratified"))
         grid_par%use_vamp_equivalences = &
              var_list_get_lval (var_list, var_str ("?use_vamp_equivalences"))
         grid_par%channel_weights_power = &
              var_list_get_rval (var_list, var_str ("channel_weights_power"))
         run_id = &
              var_list_get_sval (var_list, var_str ("$run_id"))
         rebuild_grids = &
              var_list_get_lval (var_list, var_str ("?rebuild_grids"))
       end associate
       allocate (mci_vamp_t :: mci)
       select type (mci)
       type is (mci_vamp_t)
          call mci%set_grid_parameters (grid_par)
          if (run_id /= "") then
             call mci%set_grid_filename (process_id, run_id)
          else
             call mci%set_grid_filename (process_id)
          end if
          call mci%set_rebuild_flag (rebuild_grids)
       end select
    case default
       call msg_fatal ("Integrator '" &
            // char (integration_method) // "' not implemented")
    end select
  end subroutine dispatch_mci
  
  subroutine dispatch_phs (phs, global, mapping_defaults, phs_par)
    class(phs_config_t), allocatable, intent(inout) :: phs
    type(rt_data_t), intent(in) :: global
    type(mapping_defaults_t), intent(in), optional :: mapping_defaults
    type(phs_parameters_t), intent(in), optional :: phs_par
    type(string_t) :: phs_method, phs_file
    logical :: use_equivalences
    integer :: u_phs
    logical :: exist
    phs_method = var_list_get_sval (global%var_list, &
         var_str ("$phs_method"))
    phs_file = var_list_get_sval (global%var_list, &
         var_str ("$phs_file"))
    use_equivalences = var_list_get_lval (global%var_list, &
         var_str ("?use_vamp_equivalences"))
    select case (char (phs_method))
    case ("single")
       allocate (phs_single_config_t :: phs)
    case ("wood", "default")
       allocate (phs_wood_config_t :: phs)
       select type (phs)
       type is (phs_wood_config_t)
          if (phs_file /= "") then
             inquire (file = char (phs_file), exist = exist)
             if (exist) then
                call msg_message ("Phase space: reading configuration from '" &
                     // char (phs_file) // "'")
                u_phs = free_unit ()
                open (u_phs, file = char (phs_file), &
                     action = "read", status = "old")
                call phs%set_input (u_phs)
             else
                call msg_fatal ("Phase space: configuration file '" &
                     // char (phs_file) // "' not found")
             end if
          end if
          if (present (phs_par)) &
               call phs%set_parameters (phs_par)
          if (use_equivalences) &
               call phs%enable_equivalences ()
          if (present (mapping_defaults)) &
               call phs%set_mapping_defaults (mapping_defaults)
       end select
    case default
       call msg_fatal ("Phase space: parameterization method '" &
            // char (phs_method) // "' not implemented")
    end select
  end subroutine dispatch_phs
  
  subroutine dispatch_rng (rng, global)
    class(rng_t), allocatable, intent(inout) :: rng
    type(rt_data_t), intent(in) :: global
    type(string_t) :: rng_method
    rng_method = var_list_get_sval (global%var_list, &
         var_str ("$rng_method"))
    select case (char (rng_method))
    case ("unit_test")
       allocate (rng_test_t :: rng)
    case ("tao")
       allocate (rng_tao_t :: rng)
    case default
       call msg_fatal ("Random-number generator '" &
            // char (rng_method) // "' not implemented")
    end select
  end subroutine dispatch_rng
  
  subroutine dispatch_sf_data (data, sf_method, i_beam, global)
    class(sf_data_t), allocatable, intent(inout) :: data
    type(string_t), intent(in) :: sf_method
    integer, dimension(:), intent(in) :: i_beam
    type(rt_data_t), intent(inout) :: global
    type(model_t), pointer :: model
    type(flavor_t), dimension(2) :: flv_beam
    real(default) :: isr_alpha, isr_q_max, isr_mass
    integer :: isr_order
    logical :: isr_recoil
    real(default) :: epa_alpha, epa_x_min, epa_q_min, epa_e_max, epa_mass
    logical :: epa_recoil
    real(default) :: ewa_x_min, ewa_q_min, ewa_pt_max, ewa_mass
    logical :: ewa_keep_momentum, ewa_keep_energy
    logical :: pdf_builtin_have_name
    type(string_t) :: pdf_name, pdf_path
    flv_beam = beam_data_get_flavor (global%beam_data)
    model => global%model
    associate (var_list => global%var_list)
      select case (char (sf_method))
      case ("sf_test_0", "sf_test_1")
         allocate (sf_test_data_t :: data)
         select type (data)
         type is (sf_test_data_t)
            select case (char (sf_method))
            case ("sf_test_0");  call data%init (model, flv_beam(i_beam(1)))
            case ("sf_test_1");  call data%init (model, flv_beam(i_beam(1)), &
                 mode = 1)
            end select
         end select
      case ("pdf_builtin")
         allocate (pdf_builtin_data_t :: data)
         pdf_builtin_have_name = &
              var_list_is_known (var_list, var_str ("$pdf_builtin_set"))
         if (pdf_builtin_have_name) &              
              pdf_name = var_list_get_sval &
                             (var_list, var_str ("$pdf_builtin_set"))
         pdf_builtin_have_name = trim (pdf_name) /= ""
         pdf_path = ""
         if (var_list_is_known (var_list, var_str ("$pdf_builtin_path"))) &
              pdf_path = var_list_get_sval (var_list, &
                 var_str ("$pdf_builtin_path"))
         if (trim (pdf_path) == "") & 
              pdf_path = global%os_data%pdf_builtin_datapath
         select type (data)
         type is (pdf_builtin_data_t)
            if (pdf_builtin_have_name) then
               call data%init (global%pdf_builtin_status, model, &
                    flv_beam(i_beam(1)), name = pdf_name, path = pdf_path)
            else
               call data%init (global%pdf_builtin_status, model, &
                    flv_beam(i_beam(1)), path = pdf_path)
            end if
         end select
      case ("lhapdf")
         allocate (lhapdf_data_t :: data)
         select type (data)
         type is (lhapdf_data_t)
            call data%init &
                 (global%lhapdf_status, model, flv_beam(i_beam(1)))
         end select
      case ("isr")
         allocate (isr_data_t :: data)
         isr_alpha = var_list_get_rval (var_list, var_str ("isr_alpha"))
         if (isr_alpha == 0) then
            isr_alpha = (var_list_get_rval (var_list, var_str ("ee"))) &
                 ** 2 / (4 * PI)
         end if
         isr_q_max = var_list_get_rval (var_list, var_str ("isr_q_max"))
         if (isr_q_max == 0) then
            isr_q_max = global%beam_data%sqrts
         end if
         isr_mass   = var_list_get_rval (var_list, var_str ("isr_mass"))
         isr_order  = var_list_get_ival (var_list, var_str ("isr_order"))
         isr_recoil = var_list_get_lval (var_list, var_str ("?isr_recoil")) 
         select type (data)
         type is (isr_data_t)
            call data%init &
                 (model, flv_beam (i_beam(1)), isr_alpha, isr_q_max, &
                 isr_mass, isr_order, isr_recoil)
            call data%check ()
         end select
      case ("epa")
         allocate (epa_data_t :: data)
         epa_alpha = var_list_get_rval (var_list, var_str ("epa_alpha"))
         if (epa_alpha == 0) then
            epa_alpha = (var_list_get_rval (var_list, var_str ("ee"))) &
                 ** 2 / (4 * PI)
         end if         
         epa_x_min = var_list_get_rval (var_list, var_str ("epa_x_min"))
         epa_q_min = var_list_get_rval (var_list, var_str ("epa_q_min"))
         epa_e_max = var_list_get_rval (var_list, var_str ("epa_e_max"))
         if (epa_e_max == 0) then
            epa_e_max = global%beam_data%sqrts
         end if
         epa_mass   = var_list_get_rval (var_list, var_str ("epa_mass"))
         epa_recoil = var_list_get_lval (var_list, var_str ("?epa_recoil"))
         select type (data)            
         type is (epa_data_t)
            call data%init &
                 (model, flv_beam (i_beam(1)), epa_alpha, epa_x_min, &
                 epa_q_min, epa_e_max, epa_mass, epa_recoil)
            call data%check ()
         end select
      case ("ewa")
         allocate (ewa_data_t :: data)
         ewa_x_min = var_list_get_rval (var_list, var_str ("ewa_x_min"))
         ewa_q_min = var_list_get_rval (var_list, var_str ("ewa_q_min"))
         ewa_pt_max = var_list_get_rval (var_list, var_str ("ewa_pt_max")) 
         if (ewa_pt_max == 0) then
            ewa_pt_max = global%beam_data%sqrts
         end if
         ewa_mass = var_list_get_rval (var_list, var_str ("ewa_mass"))  
         ewa_keep_momentum = var_list_get_lval (var_list, &
              var_str ("?ewa_keep_momentum"))
         ewa_keep_energy = var_list_get_lval (var_list, &
                    var_str ("?ewa_keep_energy"))                  
         if (ewa_keep_momentum .and. ewa_keep_energy) &
              call msg_fatal (" EWA cannot conserve both energy " &
                 // "and momentum.")          
         select type (data)
         type is (ewa_data_t)
            if (ewa_mass /= 0) then
               call data%init &
                    (model, flv_beam (i_beam(1)), ewa_x_min, ewa_q_min, &
                    ewa_pt_max, global%beam_data%sqrts, ewa_keep_momentum, &
                    ewa_keep_energy, ewa_mass)
            else
               call data%init &
                    (model, flv_beam (i_beam(1)), ewa_x_min, ewa_q_min, &
                    ewa_pt_max, global%beam_data%sqrts, ewa_keep_momentum, &
                    ewa_keep_energy)
            end if
            call data%check ()
         end select
      case ("energy_scan")
      case default
         call msg_bug ("Structure function '" &
              // char (sf_method) // "' not implemented yet")
      end select
    end associate
  end subroutine dispatch_sf_data
  
  function strfun_mode (name) result (n)
    type(string_t), intent(in) :: name
    integer :: n
    select case (char (name))
    case ("none")
       n = 0
    case ("sf_test_0", "sf_test_1")
       n = 1
    case ("pdf_builtin","lhapdf","isr","epa","ewa")
       n = 1
    case default
       call msg_bug ("Structure function '" // char (name) &
            // "' not supported yet")
    end select
  end function strfun_mode
    
  subroutine dispatch_sf_config (sf_config, sf_channel, global)
    type(sf_config_t), dimension(:), allocatable, intent(out) :: sf_config
    type(sf_channel_t), dimension(:), allocatable, intent(out) :: sf_channel
    type(rt_data_t), intent(inout) :: global
    type(beam_structure_t) :: beam_structure
    class(sf_data_t), allocatable :: sf_data
    integer :: n_record, i
    beam_structure = global%beam_structure
    call beam_structure%expand (strfun_mode)
    n_record = beam_structure%get_n_record ()
    allocate (sf_config (n_record))
    do i = 1, n_record
       call dispatch_sf_data (sf_data, &
            beam_structure%get_name (i), &
            beam_structure%get_i_entry (i), &
            global)
       call sf_config(i)%init (beam_structure%get_i_entry (i), sf_data)
       deallocate (sf_data)
    end do
    select case (char (beam_structure%to_string ()))
    !!! JRR: WK please check
    !!! What about this? How for the other SFs?
    !!! Single structure function not covered yet
    case ("pdf_builtin, none => none, pdf_builtin")
       allocate (sf_channel (1))
       call sf_channel(1)%init (2)
       call sf_channel(1)%set_s_mapping ([1, 2], power = 2._default)
    case ("lhapdf, none => none, lhapdf")
       allocate (sf_channel (1))
       call sf_channel(1)%init (2)
       call sf_channel(1)%set_s_mapping ([1, 2], power = 2._default)
    case ("isr, none => none, isr")
       allocate (sf_channel (1))
       call sf_channel(1)%init(2)
    case ("epa, none => none, epa")
       allocate (sf_channel (1))
       call sf_channel(1)%init(2)       
    end select
  end subroutine dispatch_sf_config
    
  subroutine dispatch_eio (eio, method, global)
    class(eio_t), intent(inout), allocatable :: eio
    type(string_t), intent(in) :: method
    type(rt_data_t), intent(in) :: global
    select case (char (method))
    case ("raw")
       allocate (eio_raw_t :: eio)
    case ("lhef")
       allocate (eio_lhef_t :: eio)
    case ("weight_stream")
       allocate (eio_weights_t :: eio)
    case default
       call msg_fatal ("Event I/O method '" // char (method) &
            // "' not implemented")
    end select
  end subroutine dispatch_eio
  

  subroutine dispatch_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (dispatch_1, "dispatch_1", &
         "process configuration method", &
         u, results)
    call test (dispatch_2, "dispatch_2", &
         "process core", &
         u, results)
    call test (dispatch_3, "dispatch_3", &
         "integration method", &
         u, results)
    call test (dispatch_4, "dispatch_4", &
         "phase-space configuration", &
         u, results)
    call test (dispatch_5, "dispatch_5", &
         "random-number generator", &
         u, results)
    call test (dispatch_6, "dispatch_6", &
         "configure phase space using file", &
         u, results)
    call test (dispatch_7, "dispatch_7", &
         "structure-function data", &
         u, results)
    call test (dispatch_8, "dispatch_8", &
         "beam structure", &
         u, results)
    call test (dispatch_9, "dispatch_9", &
         "event I/O", &
         u, results)
    call test (dispatch_10, "dispatch_10", &
         "process core update", &
         u, results)
end subroutine dispatch_test

  subroutine dispatch_1 (u)
    integer, intent(in) :: u
    type(string_t), dimension(2) :: prt_in, prt_out
    type(rt_data_t), target :: global
    class(prc_core_def_t), allocatable :: core_def
    
    write (u, "(A)")  "* Test output: dispatch_1"
    write (u, "(A)")  "*   Purpose: select process configuration method"
    write (u, "(A)")

    call global%global_init ()
    call var_list_set_log (global%var_list, var_str ("?omega_openmp"), &
         .false., is_known = .true.)

    prt_in = [var_str ("a"), var_str ("b")]
    prt_out = [var_str ("c"), var_str ("d")]

    write (u, "(A)")  "* Allocate core_def as prc_test_def"

    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call dispatch_core_def (core_def, prt_in, prt_out, global)
    select type (core_def)
    type is (prc_test_def_t)
       call core_def%write (u)
    end select
    
    deallocate (core_def)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate core_def as omega_def"
    write (u, "(A)")

    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("omega"), is_known = .true.)
    call dispatch_core_def (core_def, prt_in, prt_out, global)
    select type (core_def)
    type is (omega_def_t)
       call core_def%write (u)
    end select
    
    call global%final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_1"
    
  end subroutine dispatch_1
  
  subroutine dispatch_2 (u)
    integer, intent(in) :: u
    type(string_t), dimension(2) :: prt_in, prt_out
    type(rt_data_t), target :: global
    class(prc_core_def_t), allocatable :: core_def
    class(prc_core_t), allocatable :: core
    type(model_t), pointer :: model
    
    write (u, "(A)")  "* Test output: dispatch_2"
    write (u, "(A)")  "*   Purpose: select process configuration method"
    write (u, "(A)")  "             and allocate process core"
    write (u, "(A)")

    call syntax_model_file_init ()
    call global%global_init ()

    prt_in = [var_str ("a"), var_str ("b")]
    prt_out = [var_str ("c"), var_str ("d")]

    write (u, "(A)")  "* Allocate core as test_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("unit_test"), is_known = .true.)
    call dispatch_core_def (core_def, prt_in, prt_out, global)
    call dispatch_core (core, core_def)
    select type (core)
    type is (test_t)
       call core%write (u)
    end select
    
    deallocate (core)
    deallocate (core_def)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate core as prc_omega_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("omega"), is_known = .true.)
    call dispatch_core_def (core_def, prt_in, prt_out, global)

    call model_list_read_model (var_str ("Test"), var_str ("Test.mdl"), &
         global%os_data, model)

    call var_list_set_log (global%var_list, &
         var_str ("?helicity_selection_active"), &
         .true., is_known = .true.)
    call var_list_set_real (global%var_list, &
         var_str ("helicity_selection_threshold"), &
         1e9_default, is_known = .true.)
    call var_list_set_int (global%var_list, &
         var_str ("helicity_selection_cutoff"), &
         10, is_known = .true.)
    
    call dispatch_core (core, core_def, model, global%get_helicity_selection ())
    call core_def%allocate_driver (core%driver, var_str (""))

    select type (core)
    type is (prc_omega_t)
       call core%write (u)
    end select
    
    
    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_2"
    
  end subroutine dispatch_2
  
  subroutine dispatch_3 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    class(mci_t), allocatable :: mci
    type(string_t) :: process_id
    
    write (u, "(A)")  "* Test output: dispatch_3"
    write (u, "(A)")  "*   Purpose: select integration method"
    write (u, "(A)")

    call global%global_init ()
    process_id = "dispatch_3"

    write (u, "(A)")  "* Allocate MCI as midpoint_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$integration_method"), &
         var_str ("midpoint"), is_known = .true.)
    call dispatch_mci (mci, global, process_id)
    select type (mci)
    type is (mci_midpoint_t)
       call mci%write (u)
    end select

    call mci%final ()
    deallocate (mci)
    
    write (u, "(A)")
    write (u, "(A)")  "* Allocate MCI as vamp_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$integration_method"), &
         var_str ("vamp"), is_known = .true.)
    call var_list_set_int (global%var_list, var_str ("threshold_calls"), &
         1, is_known = .true.)
    call var_list_set_int (global%var_list, var_str ("min_calls_per_channel"), &
         2, is_known = .true.)
    call var_list_set_int (global%var_list, var_str ("min_calls_per_bin"), &
         3, is_known = .true.)
    call var_list_set_int (global%var_list, var_str ("min_bins"), &
         4, is_known = .true.)
    call var_list_set_int (global%var_list, var_str ("max_bins"), &
         5, is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?stratified"), &
         .false., is_known = .true.)
    call var_list_set_log (global%var_list, var_str ("?use_vamp_equivalences"),&
         .false., is_known = .true.)
    call var_list_set_real (global%var_list, var_str ("channel_weights_power"),&
         4._default, is_known = .true.)

    call dispatch_mci (mci, global, process_id)
    select type (mci)
    type is (mci_vamp_t)
       call mci%write (u)
    end select

    call mci%final ()
    deallocate (mci)
    
    call global%final ()

    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_3"
    
  end subroutine dispatch_3
  
  subroutine dispatch_4 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    class(phs_config_t), allocatable :: phs
    
    write (u, "(A)")  "* Test output: dispatch_4"
    write (u, "(A)")  "*   Purpose: select phase-space configuration method"
    write (u, "(A)")

    call global%global_init ()

    write (u, "(A)")  "* Allocate PHS as phs_single_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$phs_method"), &
         var_str ("single"), is_known = .true.)
    call dispatch_phs (phs, global)
    call phs%write (u)

    call phs%final ()
    deallocate (phs)
    
    write (u, "(A)")
    write (u, "(A)")  "* Allocate PHS as phs_wood_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$phs_method"), &
         var_str ("wood"), is_known = .true.)
    call dispatch_phs (phs, global)
    call phs%write (u)

    call phs%final ()

    call global%final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_4"
    
  end subroutine dispatch_4
  
  subroutine dispatch_5 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    class(rng_t), allocatable :: rng
    
    write (u, "(A)")  "* Test output: dispatch_5"
    write (u, "(A)")  "*   Purpose: select random-number generator"
    write (u, "(A)")

    call global%global_init ()

    write (u, "(A)")  "* Allocate RNG as rng_test_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$rng_method"), &
         var_str ("unit_test"), is_known = .true.)
    call dispatch_rng (rng, global)
    select type (rng)
    type is (rng_test_t)
       write (u, "(3x,A)")  "rng is rng_test_t"
    end select

    deallocate (rng)
    
    write (u, "(A)")
    write (u, "(A)")  "* Allocate RNG as rng_tao_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$rng_method"), &
         var_str ("tao"), is_known = .true.)
    call dispatch_rng (rng, global)
    select type (rng)
    type is (rng_tao_t)
       write (u, "(3x,A)")  "rng is rng_tao_t"
    end select
    
    call global%final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_5"
    
  end subroutine dispatch_5
  
  subroutine dispatch_6 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    type(model_t), pointer :: model
    type(os_data_t) :: os_data
    type(process_constants_t) :: process_data
    class(phs_config_t), allocatable :: phs
    integer :: u_phs
    
    write (u, "(A)")  "* Test output: dispatch_6"
    write (u, "(A)")  "*   Purpose: select 'wood' phase-space &
         &for a test process"
    write (u, "(A)")  "*            and read phs configuration from file"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize a process"
    write (u, "(A)")

    call global%global_init ()

    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list_read_model (var_str ("Test"), &
         var_str ("Test.mdl"), os_data, model)

    call syntax_phs_forest_init ()
    
    call init_test_process_data (var_str ("dispatch_6"), process_data)

    write (u, "(A)")  "* Write phase-space file"

    u_phs = free_unit ()
    open (u_phs, file = "dispatch_6.phs", action = "write", status = "replace")
    call write_test_phs_file (u_phs, var_str ("dispatch_6"))
    close (u_phs)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate PHS as phs_wood_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, &
         var_str ("$phs_method"), &
         var_str ("wood"), is_known = .true.)
    call var_list_set_string (global%var_list, &
         var_str ("$phs_file"), &
         var_str ("dispatch_6.phs"), is_known = .true.)
    call dispatch_phs (phs, global)

    call phs%init (process_data)
    call phs%configure (sqrts = 1000._default)

    call phs%write (u)
    write (u, "(A)")
    select type (phs)
    type is (phs_wood_config_t)
       call phs%write_forest (u)
    end select

    call phs%final ()

    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_6"
    
  end subroutine dispatch_6
  
  subroutine dispatch_7 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    type(os_data_t) :: os_data
    type(flavor_t) :: flv
    type(string_t) :: sf_method
    class(sf_data_t), allocatable :: data
    
    write (u, "(A)")  "* Test output: dispatch_7"
    write (u, "(A)")  "*   Purpose: select and configure &
         &structure function data"
    write (u, "(A)")

    call global%global_init ()
    call var_list_set_string (global%var_list, var_str ("$pdf_builtin_path"), &
         var_str ("."), is_known = .true.)
    
    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list_read_model (var_str ("QCD"), &
         var_str ("QCD.mdl"), os_data, global%model)
    
    call flavor_init (flv, PROTON, global%model)

    call reset_interaction_counter ()
    call beam_data_init_sqrts (global%beam_data, &
         14000._default, [flv, flv])
    
    write (u, "(A)")  "* Allocate data as sf_pdf_builtin_t"
    write (u, "(A)")

    sf_method = "pdf_builtin"
    call dispatch_sf_data (data, sf_method, [1], global)
    call data%write (u)

    write (u, "(A)")
    write (u, "(1x,A,99(1x,I0))")  "PDG(out) = ", data%get_pdg_out ()

    deallocate (data)
    
    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_7"
    
  end subroutine dispatch_7
  
  subroutine dispatch_8 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    type(os_data_t) :: os_data
    type(flavor_t) :: flv
    type(sf_config_t), dimension(:), allocatable :: sf_config
    type(sf_channel_t), dimension(:), allocatable :: sf_channel
    integer :: i
    
    write (u, "(A)")  "* Test output: dispatch_8"
    write (u, "(A)")  "*   Purpose: configure a structure-function chain"
    write (u, "(A)")

    call global%global_init ()
    call var_list_set_string (global%var_list, var_str ("$pdf_builtin_path"), &
         var_str ("."), is_known = .true.)
    
    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list_read_model (var_str ("QCD"), &
         var_str ("QCD.mdl"), os_data, global%model)
    
    write (u, "(A)")  "* Allocate LHC beams with PDF builtin"
    write (u, "(A)")

    call flavor_init (flv, PROTON, global%model)

    call reset_interaction_counter ()
    call beam_data_init_sqrts (global%beam_data, &
         14000._default, [flv, flv])
    
    call global%beam_structure%init ([1], global%var_list)
    call global%beam_structure%set_entry (1, 1, var_str ("pdf_builtin"))
    
    call dispatch_sf_config (sf_config, sf_channel, global)

    do i = 1, size (sf_config)
       call sf_config(i)%write (u)
    end do
    write (u, "(1x,A)")  "Mapping configuration:"
    do i = 1, size (sf_channel)
       write (u, "(2x)", advance = "no")
       call sf_channel(i)%write (u)
    end do

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_8"
    
  end subroutine dispatch_8
  
  subroutine dispatch_9 (u)
    integer, intent(in) :: u
    type(rt_data_t), target :: global
    class(eio_t), allocatable :: eio
    
    write (u, "(A)")  "* Test output: dispatch_9"
    write (u, "(A)")  "*   Purpose: allocate an event I/O (eio) stream"
    write (u, "(A)")

    call global%global_init ()
    
    write (u, "(A)")  "* Allocate as raw"
    write (u, "(A)")
    
    call dispatch_eio (eio, var_str ("raw"), global)

    call eio%write (u)

    call eio%final ()
    deallocate (eio)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate as LHEF:"
    write (u, "(A)")
    
    call dispatch_eio (eio, var_str ("lhef"), global)

    call eio%write (u)

    call eio%final ()
    deallocate (eio)

    write (u, "(A)")
    write (u, "(A)")  "* Allocate as weight_stream"
    write (u, "(A)")
    
    call dispatch_eio (eio, var_str ("weight_stream"), global)

    call eio%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"

    call eio%final ()
    call global%final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_9"
    
  end subroutine dispatch_9
  
  subroutine dispatch_10 (u)
    integer, intent(in) :: u
    type(string_t), dimension(2) :: prt_in, prt_out
    type(rt_data_t), target :: global
    class(prc_core_def_t), allocatable :: core_def
    class(prc_core_t), allocatable :: core, saved_core
    type(model_t), pointer :: model
    type(var_list_t), pointer :: var_list
    
    write (u, "(A)")  "* Test output: dispatch_10"
    write (u, "(A)")  "*   Purpose: select process configuration method,"
    write (u, "(A)")  "             allocate process core,"
    write (u, "(A)")  "             temporarily reset parameters"
    write (u, "(A)")

    call syntax_model_file_init ()
    call global%global_init ()

    prt_in = [var_str ("a"), var_str ("b")]
    prt_out = [var_str ("c"), var_str ("d")]

    write (u, "(A)")  "* Allocate core as prc_omega_t"
    write (u, "(A)")

    call var_list_set_string (global%var_list, var_str ("$method"), &
         var_str ("omega"), is_known = .true.)
    call dispatch_core_def (core_def, prt_in, prt_out, global)

    call model_list_read_model (var_str ("Test"), var_str ("Test.mdl"), &
         global%os_data, model)

    call dispatch_core (core, core_def, model)
    call core_def%allocate_driver (core%driver, var_str (""))

    select type (core)
    type is (prc_omega_t)
       call core%write (u)
    end select
    
    write (u, "(A)")
    write (u, "(A)")  "* Update core with modified model and helicity selection"
    write (u, "(A)")

    var_list => model_get_var_list_ptr (model)
    call var_list_set_real (var_list, var_str ("gy"), 2._default, &
         is_known = .true.)
    call model_parameters_update (model)

    call var_list_set_log (global%var_list, &
         var_str ("?helicity_selection_active"), &
         .true., is_known = .true.)
    call var_list_set_real (global%var_list, &
         var_str ("helicity_selection_threshold"), &
         2e10_default, is_known = .true.)
    call var_list_set_int (global%var_list, &
         var_str ("helicity_selection_cutoff"), &
         5, is_known = .true.)
    
    call dispatch_core_update (core, model, global%get_helicity_selection (), &
         saved_core)
    select type (core)
    type is (prc_omega_t)
       call core%write (u)
    end select
    
    write (u, "(A)")
    write (u, "(A)")  "* Restore core from save"
    write (u, "(A)")

    call dispatch_core_restore (core, saved_core)
    select type (core)
    type is (prc_omega_t)
       call core%write (u)
    end select
    
    call global%final ()
    call model_list_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: dispatch_10"
    
  end subroutine dispatch_10
  

end module dispatch
