! WHIZARD 2.2.0 alpha Sep 12 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module eio_lhef
  
  use kinds !NODEP!
  use file_utils !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use diagnostics !NODEP!
  use unit_tests

  use lorentz !NODEP!
  use particles
  use beams
  use processes
  use events
  use eio_base
  use hep_common

  implicit none
  private

  public :: eio_lhef_t
  public :: eio_lhef_test

  type, extends (eio_t) :: eio_lhef_t
     type(string_t) :: filename
     logical :: writing = .false.
     integer :: unit = 0
   contains
     procedure :: write => eio_lhef_write
     procedure :: final => eio_lhef_final
     procedure :: init_out => eio_lhef_init_out
     procedure :: init_in => eio_lhef_init_in
     procedure :: switch_inout => eio_lhef_switch_inout
     procedure :: output => eio_lhef_output
     procedure :: input_i_prc => eio_lhef_input_i_prc
     procedure :: input_event => eio_lhef_input_event
  end type eio_lhef_t
  

contains
  
  subroutine eio_lhef_write (object, unit)
    class(eio_lhef_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    write (u, "(1x,A)")  "LHEF event stream:"
    if (object%writing) then
       write (u, "(3x,A,A)")  "Writing to file   = ", char (object%filename)
    else
       write (u, "(3x,A)")  "[closed]"
    end if
  end subroutine eio_lhef_write
  
  subroutine eio_lhef_final (object)
    class(eio_lhef_t), intent(inout) :: object
    if (object%writing) then
       write (msg_buffer, "(A,A,A)")  "Events: closing LHEF file '", &
            char (object%filename), "'"
       call msg_message ()
       call lhef_write_footer (object%unit)
       close (object%unit)
       object%writing = .false.
    end if
  end subroutine eio_lhef_final
  
  subroutine eio_lhef_init_out (eio, sample, process_ptr, data, md5sum, success)
    class(eio_lhef_t), intent(inout) :: eio
    type(string_t), intent(in) :: sample
    type(process_ptr_t), dimension(:), intent(in) :: process_ptr
    type(event_sample_data_t), intent(in), optional :: data
    character(32), intent(in), optional :: md5sum
    logical, intent(out), optional :: success
    integer :: i
    if (.not. present (data)) &
         call msg_bug ("LHEF initialization: missing data")
    if (data%n_beam /= 2) &
         call msg_fatal ("LHEF: defined for scattering processes only")
    eio%filename = sample // ".lhef"
    eio%unit = free_unit ()
    write (msg_buffer, "(A,A,A)")  "Events: writing to LHEF file '", &
         char (eio%filename), "'"
    call msg_message ()
    eio%writing = .true.
    open (eio%unit, file = char (eio%filename), &
         action = "write", status = "replace")
    call lhef_write_header (eio%unit)
    call heprup_init &
         (data%pdg_beam, &
          data%energy_beam, &
          n_processes = data%n_proc, &
          unweighted = data%unweighted, &
          negative_weights = data%negative_weights)             
    do i = 1, data%n_proc
       call heprup_set_process_parameters (i = i, &
            process_id = data%proc_num_id(i), &
            cross_section = data%cross_section(i), &
            error = data%error(i))
    end do
    call heprup_write_lhef (eio%unit)
    if (present (success))  success = .true.
  end subroutine eio_lhef_init_out
    
  subroutine eio_lhef_init_in (eio, sample, process_ptr, md5sum, success)
    class(eio_lhef_t), intent(inout) :: eio
    type(string_t), intent(in) :: sample
    type(process_ptr_t), dimension(:), intent(in) :: process_ptr
    character(32), intent(in), optional :: md5sum
    logical, intent(out), optional :: success
    call msg_bug ("LHEF: event input not supported")
  end subroutine eio_lhef_init_in
    
  subroutine eio_lhef_switch_inout (eio, success)
    class(eio_lhef_t), intent(inout) :: eio
    logical, intent(out), optional :: success
    call msg_bug ("LHEF: in-out switch not supported")
  end subroutine eio_lhef_switch_inout
  
  subroutine eio_lhef_output (eio, event, i_prc)
    class(eio_lhef_t), intent(inout) :: eio
    type(event_t), intent(in), target :: event
    integer, intent(in) :: i_prc
    real(default) :: weight, sqme_evt, sqme_prc
    if (eio%writing) then
       call hepeup_from_event (event, process_index = i_prc)
       call hepeup_write_lhef (eio%unit)
    else
       call eio%write ()
       call msg_fatal ("LHEF file is not open for writing")
    end if
  end subroutine eio_lhef_output

  subroutine eio_lhef_input_i_prc (eio, i_prc, iostat)
    class(eio_lhef_t), intent(inout) :: eio
    integer, intent(out) :: i_prc
    integer, intent(out) :: iostat
    call msg_bug ("LHEF: event input not supported")
  end subroutine eio_lhef_input_i_prc

  subroutine eio_lhef_input_event (eio, event, iostat)
    class(eio_lhef_t), intent(inout) :: eio
    type(event_t), intent(inout), target :: event
    integer, intent(out) :: iostat
    call msg_bug ("LHEF: event input not supported")
  end subroutine eio_lhef_input_event

  subroutine lhef_write_header (unit)
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit);  if (u < 0)  return
    write (u, '(A)') '<LesHouchesEvents version="1.0">'
    write (u, '(A)') '<header>'
    write (u, '(A)') '  <generator_name>WHIZARD</generator_name>'
    write (u, '(A)') '  <generator_version>2.2.0 alpha</generator_version>'
    write (u, '(A)') '</header>'
  end subroutine lhef_write_header

  subroutine lhef_write_footer (unit)
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit);  if (u < 0)  return
    write (u, '(A)') '</LesHouchesEvents>'
  end subroutine lhef_write_footer


  subroutine eio_lhef_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    call test (eio_lhef_1, "eio_lhef_1", &
         "read and write event contents", &
         u, results)
  end subroutine eio_lhef_test
  
  subroutine eio_lhef_1 (u)
    integer, intent(in) :: u
    type(event_t), allocatable, target :: event
    type(process_t), allocatable, target :: process
    type(process_ptr_t) :: process_ptr
    type(process_instance_t), allocatable, target :: process_instance
    type(event_sample_data_t) :: data
    class(eio_t), allocatable :: eio
    type(string_t) :: sample
    integer :: u_file, iostat
    character(80) :: buffer

    write (u, "(A)")  "* Test output: eio_lhef_1"
    write (u, "(A)")  "*   Purpose: generate an event and write weight to file"
    write (u, "(A)")

    write (u, "(A)")  "* Initialize test process"
 
    allocate (process)
    process_ptr%ptr => process
    allocate (process_instance)
    call prepare_test_process (process, process_instance)
    call process_instance%setup_event_data ()
 
    allocate (event)
    call event%connect (process_instance)
    
    call data%init (1)
    data%n_beam = 2
    data%pdg_beam = 25
    data%energy_beam = 500
    data%cross_section(1) = 100
    data%error(1) = 1

    write (u, "(A)")
    write (u, "(A)")  "* Generate and write an event"
    write (u, "(A)")
 
    sample = "eio_lhef_1"
 
    allocate (eio_lhef_t :: eio)
    
    call eio%init_out (sample, [process_ptr], data)
    call event%generate (1, [0._default, 0._default])

    call eio%output (event, i_prc = 42)
    call eio%write (u)
    call eio%final ()

    write (u, "(A)")
    write (u, "(A)")  "* File contents:"
    write (u, "(A)")

    open (u_file, file = "eio_lhef_1.lhef", &
         action = "read", status = "old")
    do
       read (u_file, "(A)", iostat = iostat)  buffer
       if (iostat /= 0)  exit
       write (u, "(A)") trim (buffer)
    end do
    close (u_file)
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"
 
    call event%final ()
    deallocate (event)
 
    call cleanup_test_process (process, process_instance)
    deallocate (process_instance)
    deallocate (process)
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: eio_lhef_1"
    
  end subroutine eio_lhef_1
  

end module eio_lhef
