! WHIZARD 2.2.0 alpha Sep 12 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module qcd_parameters

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
!!! JRR: WK please check
!!! Did you intend to move these integer variables
!!! defined now in pdf_builtin.f90 over to limits.f90. 
!  use limits, only: CTEQ6L !NODEP!
  use system_dependencies !NODEP!
  use constants !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use sm_physics !NODEP!
  use pdf_builtin !NODEP!
  use md5
  use variables
  use sf_lhapdf
  
  implicit none
  private

  public :: qcd_parameters_t

  type :: qcd_parameters_t
     logical :: alpha_s_is_fixed = .true.
     integer :: order = 0
     integer :: nf = 0
     logical :: alpha_s_from_mz = .true.
     logical :: mz_is_known = .false.
     real(default) :: mz = 0
     logical :: alpha_s_mz_is_known = .false.
     real(default) :: alpha_s_mz = 0
     real(default) :: lambda = 0
     real(default) :: alpha_s_at_scale = 0
     integer :: pdf_builtin_set = CTEQ6L
     logical :: alpha_s_from_lhapdf = .false.
     type(string_t) :: lhapdf_dir
     type(string_t) :: lhapdf_file
     integer :: lhapdf_set = 0
     integer :: lhapdf_member = 0
     integer :: lhapdf_photon_scheme = 0
     logical :: alpha_s_from_pdf_builtin = .false.
   contains
     procedure :: init => qcd_parameters_basic_init
     procedure :: init_pdf_builtin => qcd_parameters_init_pdf_builtin
     procedure :: init_lhapdf => qcd_parameters_init_lhapdf
     procedure :: write => qcd_parameters_write
     procedure ::  get_md5sum => qcd_parameters_get_md5sum
     procedure :: update_alpha_s => qcd_parameters_update_alpha_s
  end type qcd_parameters_t


  interface
     double precision function alphasPDF (Q)
        double precision, intent(in) :: Q
     end function alphasPDF
  end interface

contains

  subroutine qcd_parameters_basic_init (qcd, var_list)
    class(qcd_parameters_t), intent(inout) :: qcd
    type(var_list_t), intent(in), target :: var_list
    qcd%alpha_s_is_fixed = &
         var_list_get_lval (var_list, var_str ("?alpha_s_is_fixed"))
    qcd%order = &
         var_list_get_ival (var_list, var_str ("alpha_s_order"))
    qcd%nf = &
         var_list_get_ival (var_list, var_str ("alpha_s_nf"))
    qcd%alpha_s_from_mz = &
         var_list_get_lval (var_list, var_str ("?alpha_s_from_mz"))
    qcd%alpha_s_from_lhapdf = &
         var_list_get_lval (var_list, var_str ("?alpha_s_from_lhapdf"))
    qcd%alpha_s_from_pdf_builtin = &
         var_list_get_lval (var_list, var_str ("?alpha_s_from_pdf_builtin"))    
    if (qcd%alpha_s_from_lhapdf .and. qcd%alpha_s_from_pdf_builtin) then
        call msg_fatal (" Mixing alphas evolution",  &
             (/ var_str (" from LHAPDF and builtin PDF is not permitted") /))
    end if 
    qcd%mz_is_known = &
         var_list_is_known (var_list, var_str ("mZ"))
    if (qcd%mz_is_known)  qcd%mz = &
         var_list_get_rval (var_list, var_str ("mZ"))
    qcd%alpha_s_mz_is_known = &
         var_list_is_known (var_list, var_str ("alphas"))
    if (qcd%alpha_s_mz_is_known)  qcd%alpha_s_mz = &
         var_list_get_rval (var_list, var_str ("alphas"))
    qcd%lambda = &
         var_list_get_rval (var_list, var_str ("lambda_qcd"))
  end subroutine qcd_parameters_basic_init

  subroutine qcd_parameters_init_pdf_builtin (qcd, pdf_builtin_status, &
       pdf_builtin_datapath, pdf_builtin_set, var_list)
    class(qcd_parameters_t), intent(inout) :: qcd
    type(pdf_builtin_status_t), intent(inout) :: pdf_builtin_status
    type(string_t), intent(in) :: pdf_builtin_datapath
    type(string_t), intent(in), optional :: pdf_builtin_set
    type(var_list_t), intent(in), optional :: var_list
    type(string_t) :: name
    if (qcd%alpha_s_from_pdf_builtin) then
       if (present (pdf_builtin_set)) then
          name = pdf_builtin_set
       else
          name = var_list_get_sval (var_list, &
               var_str ("$pdf_builtin_set"))  ! $
       end if
       call msg_message ("Built-in PDF set used for alpha_s = " &
            // '"' // char (name) // '"')
       qcd%pdf_builtin_set = pdf_get_id (name)
       call pdf_init (pdf_builtin_status, qcd%pdf_builtin_set, &
            pdf_builtin_datapath)
    end if
  end subroutine qcd_parameters_init_pdf_builtin

  subroutine qcd_parameters_init_lhapdf (qcd, lhapdf_status, &
       lhapdf_dir, lhapdf_file, lhapdf_member, &
       var_list)
    class(qcd_parameters_t), intent(inout) :: qcd
    type(lhapdf_status_t), intent(inout) :: lhapdf_status
    type(string_t), intent(in), optional :: lhapdf_dir, lhapdf_file
    integer, intent(in), optional :: lhapdf_member
    type(var_list_t), intent(in), optional :: var_list
    if (qcd%alpha_s_from_lhapdf) then
       if (LHAPDF_AVAILABLE) then
          if (present (lhapdf_dir)) then
             qcd%lhapdf_dir = lhapdf_dir
             qcd%lhapdf_file = lhapdf_file
             qcd%lhapdf_member = lhapdf_member
          else
             qcd%lhapdf_dir = var_list_get_sval (var_list, &
                  var_str ("$lhapdf_dir"))  ! $
             qcd%lhapdf_file = var_list_get_sval (var_list, &
                  var_str ("$lhapdf_file"))  ! $
             qcd%lhapdf_member = var_list_get_ival (var_list, &
                  var_str ("lhapdf_member"))
          end if
          qcd%lhapdf_set = 1
          call msg_message ("LHAPDF set used for alpha_s = " &
               // '"' // char (qcd%lhapdf_file) // '"')
          call lhapdf_initialize (lhapdf_status, &
               qcd%lhapdf_set, qcd%lhapdf_dir, qcd%lhapdf_file, &
               qcd%lhapdf_member)
       else             
          call msg_error &
               ("LHAPDF not linked: resetting alpha_s_from_lhapdf to false")
          qcd%alpha_s_from_lhapdf = .false.
       end if
    end if
  end subroutine qcd_parameters_init_lhapdf

  subroutine qcd_parameters_write (qcd, unit)
    class(qcd_parameters_t), intent(in) :: qcd
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    write (u, *)  "QCD coupling parameters ="
    write (u, *)  "  alpha-s is fixed = ", qcd%alpha_s_is_fixed
    if (.not. qcd%alpha_s_is_fixed) then
       if (qcd%alpha_s_from_lhapdf) then
          write (u, *)  "  alpha-s from LHAPDF"
          write (u, *)  "  PDF group        = ", qcd%lhapdf_set
          write (u, *)  "  PDF member       = ", qcd%lhapdf_member
       else 
          if (qcd%alpha_s_from_pdf_builtin) then
             write (u, *)  "  alpha-s from builtin PDF"
             write (u, *)  "  PDF group        = ", qcd%pdf_builtin_set
          else
             write (u, *)  "  LLA order        = ", qcd%order
             write (u, *)  "  active flavors   = ", qcd%nf
             write (u, *)  "  use alpha-s (mZ) = ", qcd%alpha_s_from_mz
             if (qcd%alpha_s_from_mz) then
                write (u, *)  "  mZ is known      = ", qcd%mz_is_known
                if (qcd%mz_is_known) then
                   write (u, *)  "  mZ               = ", qcd%mz
                end if
             write (u, *)  "  as(mZ) is known  = ", qcd%alpha_s_mz_is_known
             if (qcd%alpha_s_mz_is_known) then
                write (u, *)  "  alpha-s (mZ)     = ", qcd%alpha_s_mz
             end if
          else
             write (u, *)  "  Lambda_QCD       = ", qcd%lambda
          end if
       end if
    end if
    write (u, *)  "  alpha-s (scale)  = ", qcd%alpha_s_at_scale
    end if
  end subroutine qcd_parameters_write

  function qcd_parameters_get_md5sum (qcd) result (md5)
    !!! JRR: WK please check
    !!! Shall md5sum be a derived type of qcd_parameters_t?
    character(32) :: md5
    class(qcd_parameters_t), intent(in) :: qcd
    integer :: u
    u = free_unit ()
    open (unit=u, status="scratch")
    call qcd_parameters_write (qcd, u)
    rewind (u)
    md5 = md5sum (u)
    close (u)
  end function qcd_parameters_get_md5sum

  subroutine qcd_parameters_update_alpha_s (qcd, scale)
    class(qcd_parameters_t), intent(inout) :: qcd
    real(default), intent(in) :: scale
    real(default) :: alpha_s
    if (.not. qcd%alpha_s_is_fixed) then
       if (qcd%alpha_s_from_lhapdf) then
          alpha_s = alphasPDF (dble (scale))          
       else 
          if (qcd%alpha_s_from_pdf_builtin) then
             alpha_s = pdf_alphas (qcd%pdf_builtin_set, scale)
          else
             if (qcd%alpha_s_from_mz) then
                if (qcd%alpha_s_mz_is_known) then
                   if (qcd%mz_is_known) then
                      alpha_s = running_as (scale, &
                           al_mz = qcd%alpha_s_mz, &
                           mz = qcd%mz, &
                           order = qcd%order, &
                           nf = real (qcd%nf, default))
                   else
                      alpha_s = running_as (scale, &
                           al_mz = qcd%alpha_s_mz, &
                           order = qcd%order, &
                           nf = real (qcd%nf, default))
                   end if
                else
                   if (qcd%mz_is_known) then
                      alpha_s = running_as (scale, &
                           mz = qcd%mz, &
                           order = qcd%order, &
                           nf = real (qcd%nf, default))
                   else
                      alpha_s = running_as (scale, &
                           order = qcd%order, &
                           nf = real (qcd%nf, default))
                   end if
                end if
             else
                alpha_s = running_as_lam (real (qcd%nf, default), scale, &
                     lambda_qcd = qcd%lambda, &
                     order = qcd%order)
             end if
          end if
       end if
       qcd%alpha_s_at_scale = alpha_s
    end if
  end subroutine qcd_parameters_update_alpha_s


end module qcd_parameters
